package dev.pfaff.altargui.pathfinder.fontkit;

import dev.pfaff.altargui.libimage.ColorType;
import dev.pfaff.altargui.libimage.ImageData;
import dev.pfaff.altargui.texture.LoadableTexture;
import dev.pfaff.altargui.types.Color;
import dev.pfaff.altargui.util.GLThread;

import java.io.IOException;
import java.lang.foreign.Arena;
import java.lang.foreign.MemorySegment;
import java.lang.foreign.ValueLayout;
import java.nio.ByteOrder;

import static dev.pfaff.altargui.util.BasicFuture.executeOn;
import static java.lang.foreign.ValueLayout.JAVA_INT;

public record FKGlyphLoadableTexture(FKFontRef font, int codepoint) implements LoadableTexture {
	private static final ValueLayout.OfInt JAVA_INT_BIG_ENDIAN = ValueLayout.JAVA_INT.withOrder(
			ByteOrder.BIG_ENDIAN);

	private static final int SIZE = 64;
	public static final FKCanvasRasterizationOptions RASTERIZATION_OPTIONS = FKCanvasRasterizationOptions.SubpixelAa;

	static final Factors FACTORS = new Factors(0.0f, 0.0f, 0.0f, 0.0f);

	record Factors(float originX, float originY, float width, float height) {
	}

	@Override
	public ImageData load() throws IOException {
		var buf = Arena.ofAuto().allocate(JAVA_INT, SIZE * SIZE);
		rasterizeCodepoint(this.font, this.codepoint, buf, SIZE, SIZE);
		unpremultiplyAlpha(buf);
		return new ImageData(SIZE, SIZE, ColorType.Rgba8, buf);
	}

	public static void rasterizeCodepoint(FKFontRef font, int codepoint, MemorySegment buf, int canvasWidth, int canvasHeight) throws IOException {
		// inspection is misguided
		//noinspection DataFlowIssue
		var idOrNull = executeOn(() -> font.glyphForCodepoint(codepoint), GLThread::submitToGLThread);
		// inspection is wrong
		//noinspection ConstantValue
		if (idOrNull == null) throw new IOException("The font does not contain the specified codepoint");
		int id = idOrNull;
		var rasterBounds = executeOn(() -> font.rasterBounds(id, SIZE, RASTERIZATION_OPTIONS), GLThread::submitToGLThread);
		var translationX = rasterBounds.originX() * FACTORS.originX + rasterBounds.width() * FACTORS.width;
//		var translationY = rasterBounds.originY() * this.factors.originY + rasterBounds.height() * this.factors.height;
		var translationY = SIZE - (rasterBounds.originY() * FACTORS.originY + rasterBounds.height() * FACTORS.height);
		executeOn(() -> font.rasterizeCodepoint(buf,
				canvasWidth,
				canvasHeight,
				FKCanvasFormat.Rgba32,
				id,
				SIZE,
				translationX,
				translationY,
				RASTERIZATION_OPTIONS), GLThread::submitToGLThread);
	}

	public static void unpremultiplyAlpha(MemorySegment buf) {
		for (long i = 0; i < buf.byteSize(); i += JAVA_INT_BIG_ENDIAN.byteSize()) {
			buf.set(JAVA_INT_BIG_ENDIAN, i, Color.fromRgba(buf.get(JAVA_INT_BIG_ENDIAN, i)).unpremultiply().rgba());
		}
	}
}
